use std::fs::{self, File};
use std::io::{self, Write};
use std::path::Path;

use tar::{Archive, Builder, EntryType, Header};

mod compression;

pub use crate::compression::CompressionType;
use crate::compression::{Encoder, Decoder};


pub fn infer_formats<P: AsRef<Path>>(archive_path: P) -> Option<CompressionType> {
    match archive_path.as_ref().to_string_lossy() {
        s if s.ends_with(".tar.gz") => Some(CompressionType::Gzip),
        s if s.ends_with(".tar.bz2") => Some(CompressionType::Bzip2),
        _ => None,
    }
}

pub fn pack<I, P, Q>(targets: I, archive_path: Q) -> io::Result<()>
where
    I: IntoIterator<Item = P>,
    P: AsRef<Path>,
    Q: AsRef<Path>,
{
    let alg = infer_formats(&archive_path).unwrap();
    pack_with(targets, archive_path, alg)
}

pub fn pack_with<I, P, Q>(targets: I, archive_path: Q, alg: CompressionType) -> io::Result<()>
where
    I: IntoIterator<Item = P>,
    P: AsRef<Path>,
    Q: AsRef<Path>,
{
    let dst = File::create(&archive_path)?;
    let encoder = Encoder::new(dst, alg);
    let mut archive = Builder::new(encoder);

    for target in targets {
        add_from_path(&mut archive, &target, &archive_path)?;
    }

    archive.finish()
}

fn add_from_path<W, P, Q>(archive: &mut Builder<W>, path: &P, archive_path: &Q) -> io::Result<()>
where
    W: Write,
    P: AsRef<Path>,
    Q: AsRef<Path>,
{
    let meta = path.as_ref().symlink_metadata()?;
    if meta.is_file() {
        archive.append_path(path)?;
    }
    else if meta.is_dir() {
        for entry in fs::read_dir(path)? {
            let child = entry?.path();
            if child == archive_path.as_ref() {
                continue;
            }
            add_from_path(archive, &child.as_path(), archive_path)?;
        }
    }
    else if meta.is_symlink() {
        let mut header = Header::new_gnu();
        header.set_entry_type(EntryType::Symlink);
        header.set_metadata(&meta);
        let target = fs::read_link(path)?;
        archive.append_link(&mut header, path, &target)?;
    }

    Ok(())
}

pub fn unpack<P, Q>(archive_path: P, output_path: Q) -> io::Result<()> 
where
    P: AsRef<Path>,
    Q: AsRef<Path>,
{
    let alg = infer_formats(&archive_path).unwrap();
    unpack_with(archive_path, output_path, alg)
}

pub fn unpack_with<P, Q>(archive_path: P, output_path: Q, alg: CompressionType) -> io::Result<()> 
where
    P: AsRef<Path>,
    Q: AsRef<Path>,
{
    let archive_file = File::open(&archive_path)?;
    let decoder = Decoder::new(archive_file, alg);
    let mut archive = Archive::new(decoder);

    if !output_path.as_ref().exists() {
        fs::create_dir(&output_path)?;
    }

    archive.unpack(&output_path)
    // check for duplicated top-level directory and fix
    // let mut iter = fs::read_dir(&output_path).into_iter().next();
    // if let (Some(entry), None) = (iter.next(), iter.next()) {
    //     let child = entry?.path();
    //     if child.file_name() == output_path.file_name() {
    //         let mut name = output_path.file_name().to_os_string();
    //         name.push("_tmp".into())
    //     }
    // }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_infer_formats() {
        assert_eq!(infer_formats("test.tar.gz"), Some(CompressionType::Gzip));
        assert_eq!(infer_formats("test.tar.bz2"), Some(CompressionType::Bzip2));
        assert_eq!(infer_formats("test.nothing"), None);
        assert_eq!(infer_formats("test"), None);
    }
}
